/*
 * Convert.c - ToAmiga Converting Functions  Version 1.04
 * Copyright (c) 1992 by L.Georgiev
 *
 * Changes since version 1.00 released:
 *  - now effects are processed even if there isn't a note info.
 *  - patterns present in sequence but not existant are ignored.
 *  - warnings when early EOF reached ( sample data truncated ).
 *  - appends identification string after last sample data byte.
 * (17-18.IV.1992)
 */

#include <stdio.h>
#include <stdlib.h>
#include "toamiga.h"

static	word	global_volume;

/*****************************************************************************
*                          Convert sample attributes                         *
******************************************************************************/
/*
 * Returns number of patterns used, number of patterns to skip,
 * and sample data length expected in a hdr_info structure (info).
 * If source is not a valid ST module, returns 0.
 */
int attrib_conv( word *source_buf, word *target_buf, struct hdr_info * info )
{
    register	int	i,j;
		char    k;
		char    st_pfx[] = "st-xx:";    /* x's will be owerwritten  */
		char    *sign_31 = "M.K.";      /* 31-instrument designator */
		char    *stm_pat,       *mod_pat;

		struct  stm_header      *src_hdr;
    register    struct  stm_instr       *source;
    register    struct  mod_instr       *target;

    src_hdr = (struct stm_header *)source_buf;
    if( src_hdr->stm_sign != 0x1A || src_hdr->file_type != 2 )
	return 0;	/* Not a valid ST module */

    for( j = 0; j < sizeof( src_hdr->module_name ); j++ )   /* Copy title */
	( (char *)target_buf)[j] = src_hdr->module_name[j];

    global_volume = src_hdr->global_volume;	/* Needed in channel_conv */

    /* Header already processed, now prepare addresses of 1st instrument */
    source = (struct stm_instr *)
	( source_buf + sizeof( struct stm_header ) / sizeof(word) );

    /* Amiga module instruments begin immediately after the title */
    target = (struct mod_instr *)
	( target_buf + sizeof(src_hdr->module_name) / sizeof(word) );

    info->data_len = 0L;		/* Inizialize sample data length */

/********************************* Main Loop ********************************/

    for( i = 0; i < SAMPLES; i++, source += sizeof( source ) / sizeof(word),
				  target += sizeof( target ) / sizeof(word)) {
	if( source->name[0] )   {   /* Name present */

	    /* Form instrument name as "st-xx:< original name >" (xx = disk)*/
	    sprintf( st_pfx + 3, "%02d:", source->disk );
	    for( j = 0; j < sizeof( st_pfx ) - 1; j++ )/* Don't copy last 0 */
		target->name[j] = st_pfx[j];
	    for( k = 0; k < sizeof( source->name ); j++, k++)
		target->name[j] = source->name[k];	/* Append original */
	}
	/* Add length of this sample rounded up at paragraph boundaries */
	info->data_len += (long)(( source->length - 1 >> 4 ) + 1 << 4 );

	source->length >>= 1;	/* Amiga length is in words, not in bytes */

	/*Save sample length & volume in reversed byte order (see toamiga.h)*/
	swab( (char *)&source->length, (char *)&target->length, 2 );
	swab( (char *)&source->volume, (char *)&target->volume, 2 );

	if( source->loop_start == 0  &&  source->loop_end == 0xFFFF )
	    continue;   /* No loop */

	/* Amiga loop start and loop end are in words, not in bytes */
	source->loop_start >>= 1;
	source->loop_end   >>= 1;

	/* In some modules (e.g. AT.STM) loop ends beyond the sample: */
	/* move both start and end back, thus keeping length the same */
	if( source->loop_end > source->length )    {
	    source->loop_start -= source->loop_end - source->length;
	    if( (signed)source->loop_start < 0 )
		source->loop_start = 0;
	    source->loop_end = source->length;
	    printf( "WARNING: Loop end value too high for \"%s\" - fixed\n",
							source->name );
	}
	/* Get loop length (= End - Start) in source->loop_end */
	source->loop_end -= source->loop_start;

	/* Store both loop start and length */
	swab( (char *)&source->loop_start, (char *)&target->loop_start, 4 );
    }

/****************** Append Pattern Sequence And Other Stuff *****************/

    stm_pat = (char *)source;
    mod_pat = (char *)target;
    mod_pat += sizeof(word);	/* Skip # of patterns & MAGIC (set later on)*/
   /*
    * Copy pattern sequence and find max pattern # used (k). Stop when number
    * of patterns saved is reached (ignore the rest of patterns in sequence
    * if any, e.g. in EYEOF.STM). This will also stop when first 'c' (0x63)
    * appears as it is always higher than number of patterns saved.
    */
    for( j = 0, k = 0;	*stm_pat < (char)src_hdr->no_of_patterns; j++)	{
	k = max( *stm_pat, k );
	*mod_pat++ = *stm_pat++;
    }
    *(word *)target = j  |  MAGIC;  /* j is the real number of patterns in */
				    /* sequence (ignore what header says) */
    for( ; j < PATTERNS; j++)
	*mod_pat++ = 0;                 /* Fill the rest space with zeroes */
    for( j = 0; j < 4; j++)
	*mod_pat++ = *sign_31++;        /* Add 31-instr. module designator */
    info->pat_used = ++k;	    /* k+1 = # patterns used in the sequence*/
   /*
    * If number of patterns saved is really greater than the max pattern #
    * in the sequence + 1, return the difference in info->pat_to_skip.
    * The caller will skip these unused patterns (e.g. in BRAIN.STM).
    */
    info->pat_to_skip = src_hdr->no_of_patterns - k;
    return 1;	/* Successful return */
}

/*******************************/
/* Convert single channel data */
/*******************************/
					    /* Returns NULL if successful or*/
char * channel_conv( word *buf )	    /* effect name if can't convert*/
{
    register    word    volume, sample, effect, param;
		word    note;
    union       channel *chan;

		char    *eff_name[]     =
   {
    "effect @", "", "", "", "", "", "", "", "", "tremolo", "",
    "volume portamento", "volume vibrato", "effect M", "effect N", "effect O",
    "volume change"
    };						  /* Nul strings never used */

		word	notab[ 10 ][ 12 ]    =	  /* Note data table */

   /*  C    Cis    D     Es    E     F    Fis    G    Gis    A     B     H  */
   {
    {0x6B0,0x650,0x5F5,0x5A0,0x54F,0x503,0x4BB,0x477,0x436,0x3FA,0x3C1,0x38B},
    {0x358,0x328,0x2FB,0x2D0,0x2A7,0x281,0x25D,0x23B,0x21B,0x1FD,0x1E0,0x1C5},
    {0x1AC,0x194,0x17D,0x168,0x154,0x141,0x12F,0x11E,0x10E, 0xFE, 0xF0, 0xE3},
    { 0xD6, 0xCA, 0xBF, 0xB4, 0xAA, 0xA0, 0x97, 0x8F, 0x87, 0x7F, 0x78, 0x71},
    { 0x6B, 0x65, 0x5F, 0x5A, 0x55, 0x50, 0x4C, 0x47, 0x43, 0x40, 0x3C, 0x39},
    { 0x35, 0x32, 0x30, 0x2D, 0x2A, 0x28, 0x26, 0x24, 0x22, 0x20, 0x1E, 0x1C},
    { 0x1B, 0x19, 0x18, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10,  0xF,  0xE},
    {  0xD,  0xD,  0xC,  0xB,  0xB,  0xA,    9,    9,    8,    8,    8,    7},
    {    7,    6,    6,    6,    5,    5,    5,    4,    4,    4,    4,    4},
    {    3,    3,    3,    3,    3,    3,    2,    2,    2,    2,    2,    2}
   };
    chan = (union channel *)buf;

/********************* Convert note, sample, and volume *********************/

    volume = chan->stm.vol_high << 3  |	 chan->stm.vol_low; /* Save volume */
    effect = chan->stm.effect;			    /* and effect data (may */
    param = chan->stm.eff_par;			    /* exist without a note */

    if( chan->stm.octave < 10	&&   chan->stm.note < 12 )  { /* Note valid */
	note = notab[ chan->stm.octave ][ chan->stm.note ];/*Fetch note data*/
	sample = chan->stm.sample;			   /* and sample # */
	chan->mod.note_data = note & 0xFF;
	chan->mod.note_high = note >> 8;	    /* Store target values */
	chan->mod.sample = sample & 0xF;
	chan->mod.sample_msb = sample >> 4;
    }
    else
	*(long *)buf = 0L; /* Clear all fields (effect one may change later)*/

    if( effect == 0  &&  param == 0 )  { /* No effect, see if volume change */
	if( volume < global_volume )  {		/* Volume change (valid for */
	    chan->mod.effect = 0xC;             /* current note only) */
	    chan->mod.eff_par = volume;
	}
	else
	    chan->mod.effect = chan->mod.eff_par = 0;	/* No effect */
	return NULL;
    }
/************************** Convert Effect Command **************************/

    chan->mod.eff_par = param;	/* Set in advance; may be changed later on */

    switch( effect )	{
	case ARPEGGIO:	/* Effect field already 0 */
            break;
        case PORT_UP:
	    chan->mod.effect = 1;
	    break;
	case PORT_DOWN:
	    chan->mod.effect = 2;
	    break;
	case NOTE_PORT:
	    chan->mod.effect = 3;
	    break;
	case VIBRATO:
	    chan->mod.effect = 4;
	    break;
	case SLIDE_VOL:
	    chan->mod.effect = 10;
	    break;
	case JUMP:
	    chan->mod.effect = 11;
	    break;
	case BREAK:
	    chan->mod.effect = 13;
	    break;
	case TEMPO:
	    chan->mod.effect = 15;
	    chan->mod.eff_par = param >> 4; /*Fast tempo converts too roughly*/
	    break;			/* but it can't be helped...  :-(  */
        case TREMOLO:
	default:
	    chan->mod.eff_par = 0;	/* Effect not recognized or */
	    return eff_name[ effect ];	/* not implemented in Amiga */
    }
    if( volume < global_volume )    {	/* Volume change has the lowest */
	effect = VOL_CHNG;		/* priority so I'm sorry but it */
	return eff_name[ effect ];	/* will have to remain the same */
    }
    return NULL;    /* Exit successfully */
}
